/*
   core/func.cc
   This file is part of the Osirion project and is distributed under
   the terms of the GNU General Public License version 2
*/

#include <map>
#include <string>
#include <iomanip>

#include "sys/sys.h"
#include "core/func.h"

namespace core
{

/* ---- Static functions for the Func registry -------------------- */

Func::Registry Func::func_registry;

Func * Func::add(const char *name, FuncPtr functionptr, unsigned int flags)
{
	Func *func = 0;
	Registry::iterator it = func_registry.find(name);
	if (it == func_registry.end()) {
		func = new Func(name, (void *)functionptr, flags & ~Func::Game);
		func_registry[std::string(name)] = func;
		//con_debug << "Function '" << name << "' registered." << std::endl;
	} else {
		con_warn << "Function '" << name << "' already registered!" << std::endl;
		func = (*it).second;
	}
	return func;
}

Func *Func::add(const char *name, GameFuncPtr gamefunctionptr, unsigned int flags)
{
	Func *func = 0;
	Registry::iterator it = func_registry.find(name);
	if (it == func_registry.end()) {
		func = new Func(name, (void *)gamefunctionptr, flags | Func::Game);
		func_registry[std::string(name)] = func;
		//con_debug << "Function '" << name << "' registered." << std::endl;
	} else {
		con_warn << "Function '" << name << "' already registered!" << std::endl;
		func = (*it).second;
	}
	return func;
}

void Func::remove(const char *name)
{
	std::map<std::string, Func *>::iterator it = func_registry.find(std::string(name));
        if (it != func_registry.end()) {
		delete (*it).second;
		func_registry.erase(it);
		//con_debug << "Function '" << name << "' unregistered." << std::endl;
	}
}

void Func::remove(const std::string &name)
{
	std::map<std::string, Func *>::iterator it = func_registry.find(name);
        if (it != func_registry.end()) {
                delete (*it).second;
                func_registry.erase(it);
		//con_debug << "Function '" << name << "' unregistered." << std::endl;
        }

}

Func *Func::find(const std::string &name)
{
	std::map<std::string, Func *>::iterator it = func_registry.find(name);
	if (it == func_registry.end())
		return 0;
	else
		return (*it).second;
}

void Func::list()
{
	Registry::iterator it;
	con_print << "Flags: G=Game S=Shared" << std::endl;

	for (it = func_registry.begin(); it != func_registry.end(); it++) {
		std::string typeindicator;
		if (((*it).second->flags() & Game) == Game)
			typeindicator += 'G';
		else
			typeindicator += ' ';
	if (((*it).second->flags() & Shared) == Shared)
			typeindicator += 'S';
		else
			typeindicator += ' ';
		con_print << " " << typeindicator <<  " " << (*it).second->name() << " " << (*it).second->info() << std::endl;
	}

	con_print << func_registry.size() << " registered functions" << std::endl;
}

/* ---- Func ------------------------------------------------------ */

Func::Func(char const * name, void *ptr, unsigned int flags)
{
	if (name)
		func_name.assign(name);
	else
		func_name.clear();

	func_info.clear();

	func_flags = flags;
	func_ptr = ptr;
}

Func::~Func()
{
	func_name.clear();
	func_ptr = 0;
	func_flags = 0;
}

void Func::set_info(const char *info)
{
	if (info)
		func_info.assign(info);
}

void Func::exec(std::string const &args)
{
	if ((flags() & Game))
		return;

	FuncPtr function = (FuncPtr) func_ptr;
	function(args);
}
	
void Func::exec(Player *player, std::string const &args)
{
	if (!(flags() & Game))
		return;
	
	GameFuncPtr gamefunction = (GameFuncPtr) func_ptr;
	gamefunction(player, args);
}

} // namespace core

